import { Component } from "@angular/core";
import {
  IonicPage,
  NavController,
  NavParams,
  LoadingController
} from "ionic-angular";
import { FormBuilder, Validators, FormGroup } from "@angular/forms";
import { TwitterApiProvider } from "../../providers/twitter-api/twitter-api";
import { Storage } from "@ionic/storage";
import { P2pStorageIpfsProvider } from "../../providers/p2p-storage-ipfs/p2p-storage-ipfs";
import { P2pDatabaseGunProvider } from "../../providers/p2p-database-gun/p2p-database-gun";
import twittertext from "twitter-text";

@IonicPage()
@Component({
  selector: "page-write-tweet",
  templateUrl: "write-tweet.html"
})
export class WriteTweetPage {
  tweet: FormGroup;
  retweetId: string;
  replyToStatusId: string;
  retweet;
  replyTweet;

  constructor(
    public navCtrl: NavController,
    public navParams: NavParams,
    private formBuilder: FormBuilder,
    private twitter: TwitterApiProvider,
    private loadingCtrl: LoadingController,
    private storage: Storage,
    private ipfs: P2pStorageIpfsProvider,
    private gun: P2pDatabaseGunProvider
  ) {
    this.retweetId = this.navParams.get("tweetId");
    this.replyToStatusId = this.navParams.get("replyToStatus");

    this.tweet = this.formBuilder.group({
      text: ["", Validators.maxLength(140)],
      p2p: [false]
    });
  }

  async ionViewDidLoad() {
    if (this.retweetId) {
      this.retweet = await this.twitter.fetchTweet(this.retweetId);
    }
    if (this.replyToStatusId) {
      this.replyTweet = await this.twitter.fetchTweet(this.replyToStatusId);
    }
  }

  get tweetCharProgress() {
    let progress = 1 - this.tweet.value["text"].length / 140;
    let radius = 8;
    let circumference = Math.PI * radius * 2;
    return progress * circumference;
  }

  async submitTweet() {
    const loading = this.loadingCtrl.create();
    loading.present();

    if (this.tweet.value.p2p) {
      const tweet = await this.buildPrivateTweet();
      const res = await this.ipfs.storeTweet(tweet);

      this.gun.storeLastTweetHashForUser(
        tweet.user_id,
        res["Hash"],
        tweet.created_at
      );

      this.gun.publishHashtags(tweet.entities.hashtags);
    } else {
      await this.twitter.tweet(
        this.tweet.value["text"],
        this.retweet,
        this.replyToStatusId
      );
    }

    loading.dismiss();
    this.navCtrl.pop();
  }

  private async buildPrivateTweet() {
    const status = this.tweet.value["text"].trim();
    const entities = await this.getEntities(status);

    return {
      full_text: status,
      user_id: await this.storage.get("userId"),
      created_at: Date.now(),
      private_tweet: true,
      display_text_range: [0, status.length],
      entities: entities
    };
  }

  private async getEntities(status: string) {
    return {
      hashtags: twittertext.extractHashtagsWithIndices(status),
      urls: twittertext.extractUrlsWithIndices(status),
      user_mentions: await this.getMentions(status)
    };
  }

  private async getMentions(status: string) {
    // extract mentions
    const entities = twittertext.extractMentionsWithIndices(status);

    // add user_id
    const entitiesWithPromises = entities.map(async mention => {
      try {
        const user = await this.twitter.fetchUserFromScreenName(
          mention.screenName
        );
        mention["id_str"] = user[0]["id_str"];
        mention["screen_name"] = mention.screenName;
        delete mention.screenName;
      } catch (err) {
        console.error(
          "There is no user signed up to twitter with username: " +
            mention.screenName
        );
      }
      return mention;
    });

    // filter for valid users and return
    return (await Promise.all(entitiesWithPromises)).filter(el =>
      el.hasOwnProperty("id_str")
    );
  }
}
